import logging
import pathlib

import pandas as pd

from nuclearmasses.io.ame_mass_file import AMEMassFile


class AMEMassParser(AMEMassFile):
    """Parse the AME mass file.

    The format is known but the provided string does not match all lines.
    We will therefore use START and END markers, which are inherited, and
    read the columns are interested in.
    """

    def __init__(self, filename: pathlib.Path, year: int):
        """Set the file to read and table year"""
        self.filename: pathlib.Path = filename
        self.year: int = year
        super().__init__(self.year)
        logging.info(f"Reading {self.filename} from {self.year}")

    def _column_names(self) -> list[str]:
        """Set the column name depending on the year"""
        match self.year:
            case _:
                return [
                    "Z",
                    "A",
                    "AMEMassExcess",
                    "AMEMassExcessError",
                    "BindingEnergyPerA",
                    "BindingEnergyPerAError",
                    "BetaDecayEnergy",
                    "BetaDecayEnergyError",
                    "AtomicNumber",
                    "AtomicMass",
                    "AtomicMassError",
                ]

    def _data_types(self) -> dict:
        """Set the data type depending on the year"""
        match self.year:
            case _:
                return {
                    "TableYear": "Int64",
                    "Symbol": "string",
                    "N": "Int64",
                    "Z": "Int64",
                    "A": "Int64",
                    "AMEMassExcess": "float64",
                    "AMEMassExcessError": "float64",
                    "BindingEnergyPerA": "float64",
                    "BindingEnergyPerAError": "float64",
                    "BetaDecayEnergy": "float64",
                    "BetaDecayEnergyError": "float64",
                    "AtomicMass": "float64",
                    "AtomicMassError": "float64",
                }

    def _na_values(self) -> dict:
        """Set the columns that have placeholder values"""
        match self.year:
            case 1983:
                return {
                    "A": [""],
                    "BetaDecayEnergy": ["", "*"],
                    "BetaDecayEnergyError": ["", "*"],
                }
            case _:
                return {
                    "BetaDecayEnergy": ["", "*"],
                    "BetaDecayEnergyError": ["", "*"],
                }

    def read_file(self) -> pd.DataFrame:
        """Read the file using it's known format

        The AMEMassFile and other functions in this class have hopefully sanitized the
        column names, data types and locations of the date so we can now make the generic
        call to parse the file.
        """
        try:
            df = pd.read_fwf(
                self.filename,
                colspecs=self.column_limits,
                names=self._column_names(),
                na_values=self._na_values(),
                keep_default_na=False,
                on_bad_lines="warn",
                skiprows=self.HEADER,
                skipfooter=self.FOOTER,
            )
            # We use the NUBASE data to define whether or not an isotope is experimentally measured,
            # so for this data we'll just drop any and all '#' characters
            df.replace("#", "", regex=True, inplace=True)

            if self.year == 1983:
                # The column headers and units are repeated in the 1983 table
                df = df[(df["A"] != "A") & (~df["AMEMassExcess"].astype("string").str.contains("keV", na=False))]
                # The A value is not in the column if it doesn't change so we need to fill down
                df["A"] = df["A"].ffill()
                # Isomeric states are sometimes included in this version of the file
                # For each row in the dataframe, if the previous row has equal A and Z, drop the current row
                df = df[~((df["A"] == df["A"].shift()) & (df["Z"] == df["Z"].shift()))]

            if self.year == 1983 or self.year == 1993 or self.year == 1995:
                df["BindingEnergyPerA"] = df["BindingEnergyPerA"].astype(float) / df["A"].astype(float)
                df["BindingEnergyPerAError"] = df["BindingEnergyPerAError"].astype(float) / df["A"].astype(float)

            df["TableYear"] = self.year
            df["N"] = pd.to_numeric(df["A"]) - pd.to_numeric(df["Z"])
            df["Symbol"] = pd.to_numeric(df["Z"]).map(self.z_to_symbol)

            # Combine the two columns to create the atomic mass then drop the redundant column
            # Pandas is happy to use '+' for any type, but mypy doesn't like it, hence the use of str.cat()
            df["AtomicMass"] = (
                df["AtomicNumber"]
                .astype("string")
                .str.cat(df["AtomicMass"].astype("string").str.replace(".", "", regex=False), sep=".")
            )
            df = df.drop(columns=["AtomicNumber"])

            # We need to rescale the error value because we combined the two columns above
            df = df.assign(AtomicMassError=df["AtomicMassError"].astype(float) / 1.0e6)
        except ValueError as e:
            print(f"Parsing error: {e}")

        return df.astype(self._data_types())



import io

import pandas as pd
import pandas.testing as pdt

from nuclearmasses.io.ame_reaction_1_parse import AMEReactionParserOne


def test_1983_rct1():
    # We are cheating a little here because this line does not have the A value in the file
    # To test properly either choose a different isotope or read from the first instance with A so it gets populated
    line = io.StringIO(
        " 186 Ir  77   15780     250      9536      20      3850     100     -7600#    300#    -2639      20    -10640#    200#"
    )
    parser = AMEReactionParserOne(line, 1983)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1983],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15780],
            "TwoNeutronSeparationEnergyError": [250],
            "TwoProtonSeparationEnergy": [9536],
            "TwoProtonSeparationEnergyError": [20],
            "QAlpha": [3850],
            "QAlphaError": [100],
            "QTwoBeta": [-7600],
            "QTwoBetaError": [300],
            "QEpsilon": [-2639],
            "QEpsilonError": [20],
            "QBetaNeutron": [-10640],
            "QBetaNeutronError": [200],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1993_rct1():
    line = io.StringIO(
        " 186 Ir  77   15618.44  270.74   9522.98   20.49   3852.98  103.94  -7419.61  145.57  -2635.85   20.03 -10622#    230#"
    )
    parser = AMEReactionParserOne(line, 1993)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1993],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15618.44],
            "TwoNeutronSeparationEnergyError": [270.74],
            "TwoProtonSeparationEnergy": [9522.89],
            "TwoProtonSeparationEnergyError": [20.49],
            "QAlpha": [3852.98],
            "QAlphaError": [103.94],
            "QTwoBeta": [-7419.61],
            "QTwoBetaError": [145.57],
            "QEpsilon": [-2635.85],
            "QEpsilonError": [20.03],
            "QBetaNeutron": [-10622],
            "QBetaNeutronError": [230],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1995_rct1():
    line = io.StringIO(
        " 186 Ir  77   15618.41  270.74   9522.89   20.49   3853.04  103.94  -7495.33  145.56  -2635.83   20.03 -10682.00  207.60"
    )
    parser = AMEReactionParserOne(line, 1995)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1995],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15618.41],
            "TwoNeutronSeparationEnergyError": [270.74],
            "TwoProtonSeparationEnergy": [9522.89],
            "TwoProtonSeparationEnergyError": [20.49],
            "QAlpha": [3853.04],
            "QAlphaError": [103.94],
            "QTwoBeta": [-7495.33],
            "QTwoBetaError": [145.56],
            "QEpsilon": [-2635.83],
            "QEpsilonError": [20.03],
            "QBetaNeutron": [-10682.00],
            "QBetaNeutronError": [207.60],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2003_rct1():
    line = io.StringIO(
        " 186 Ir  77   15704.74   32.47   9524.26   17.08   3849.65  103.31  -7458.10   26.70  -2639.77   16.57 -10561.10   44.19"
    )
    parser = AMEReactionParserOne(line, 2003)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2003],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15704.74],
            "TwoNeutronSeparationEnergyError": [32.47],
            "TwoProtonSeparationEnergy": [9524.26],
            "TwoProtonSeparationEnergyError": [17.08],
            "QAlpha": [3849.65],
            "QAlphaError": [103.31],
            "QTwoBeta": [-7458.10],
            "QTwoBetaError": [26.70],
            "QEpsilon": [-2639.77],
            "QEpsilonError": [16.57],
            "QBetaNeutron": [-10561.10],
            "QBetaNeutronError": [44.19],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2012_rct1():
    line = io.StringIO(
        " 186 Ir  77   15706.55   32.47   9527.99   17.09   3848.03  103.31  -7459.92   26.70  -2641.13   16.57 -10557.95   30.67"
    )
    parser = AMEReactionParserOne(line, 2012)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2012],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15706.55],
            "TwoNeutronSeparationEnergyError": [32.47],
            "TwoProtonSeparationEnergy": [9527.99],
            "TwoProtonSeparationEnergyError": [17.09],
            "QAlpha": [3848.03],
            "QAlphaError": [103.31],
            "QTwoBeta": [-7459.92],
            "QTwoBetaError": [26.70],
            "QEpsilon": [-2641.13],
            "QEpsilonError": [16.57],
            "QBetaNeutron": [-10557.95],
            "QBetaNeutronError": [30.67],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2016_rct1():
    line = io.StringIO(
        " 186 Ir  77   15704.13   32.47   9530.65   17.07   3848.80  103.31  -7457.49   26.70  -2642.29   16.55 -10555.52   30.67"
    )
    parser = AMEReactionParserOne(line, 2016)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2016],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15704.13],
            "TwoNeutronSeparationEnergyError": [32.47],
            "TwoProtonSeparationEnergy": [9530.65],
            "TwoProtonSeparationEnergyError": [17.07],
            "QAlpha": [3848.80],
            "QAlphaError": [103.31],
            "QTwoBeta": [-7457.49],
            "QTwoBetaError": [26.70],
            "QEpsilon": [-2642.29],
            "QEpsilonError": [16.55],
            "QBetaNeutron": [-10555.52],
            "QBetaNeutronError": [30.67],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2020_rct1():
    line = io.StringIO(
        " 186 Ir  77   15704.1312   32.4655   9530.4731   17.0698   3848.8777  103.3133  -7457.4943   26.6968  -2642.2739   16.5459 -10555.5245   30.6658"
    )
    parser = AMEReactionParserOne(line, 2020)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2020],
            "Symbol": ["Ir"],
            "A": [186],
            "Z": [77],
            "N": [109],
            "TwoNeutronSeparationEnergy": [15704.1312],
            "TwoNeutronSeparationEnergyError": [32.4655],
            "TwoProtonSeparationEnergy": [9530.4731],
            "TwoProtonSeparationEnergyError": [17.0698],
            "QAlpha": [3848.8777],
            "QAlphaError": [103.3133],
            "QTwoBeta": [-7457.4943],
            "QTwoBetaError": [26.6968],
            "QEpsilon": [-2642.2739],
            "QEpsilonError": [16.5459],
            "QBetaNeutron": [-10555.5245],
            "QBetaNeutronError": [30.6658],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)



from nuclearmasses.utils.element_converter import ElementConverter


class AMEReactionFileOne(ElementConverter):
    """Easy access to where the variables are in the first AME reaction file."""

    def __init__(self, year: int):
        """Setup the values that locate the variable."""
        super().__init__()
        match year:
            case 1983:
                self.HEADER = 30
                self.FOOTER = 0
                self.START_R1_A = 1
                self.END_R1_A = 4
                self.START_R1_Z = 8
                self.END_R1_Z = 11
                self.START_S2N = 14
                self.END_S2N = 22
                self.START_DS2N = 24
                self.END_DS2N = 30
                self.START_S2P = 32
                self.END_S2P = 39
                self.START_DS2P = 43
                self.END_DS2P = 47
                self.START_QA = 50
                self.END_QA = 57
                self.START_DQA = 60
                self.END_DQA = 65
                self.START_Q2B = 68
                self.END_Q2B = 75
                self.START_DQ2B = 78
                self.END_DQ2B = 83
                self.START_QEP = 86
                self.END_QEP = 93
                self.START_DQEP = 96
                self.END_DQEP = 101
                self.START_QBN = 103
                self.END_QBN = 111
                self.START_DQBN = 114
                self.END_DQBN = 119
            case 2020:
                self.HEADER = 35
                self.FOOTER = 0
                self.START_R1_A = 1
                self.END_R1_A = 4
                self.START_R1_Z = 8
                self.END_R1_Z = 11
                self.START_S2N = 14
                self.END_S2N = 24
                self.START_DS2N = 25
                self.END_DS2N = 34
                self.START_S2P = 36
                self.END_S2P = 46
                self.START_DS2P = 47
                self.END_DS2P = 56
                self.START_QA = 58
                self.END_QA = 68
                self.START_DQA = 69
                self.END_DQA = 78
                self.START_Q2B = 79
                self.END_Q2B = 90
                self.START_DQ2B = 91
                self.END_DQ2B = 100
                self.START_QEP = 101
                self.END_QEP = 112
                self.START_DQEP = 113
                self.END_DQEP = 122
                self.START_QBN = 123
                self.END_QBN = 134
                self.START_DQBN = 135
                self.END_DQBN = 144
            case _:
                match year:
                    case 1995 | 2003 | 2012 | 2016:
                        self.HEADER = 39
                    case 1993:
                        self.HEADER = 40
                self.FOOTER = 0
                self.START_R1_A = 1
                self.END_R1_A = 4
                self.START_R1_Z = 8
                self.END_R1_Z = 11
                self.START_S2N = 14
                self.END_S2N = 22
                self.START_DS2N = 23
                self.END_DS2N = 30
                self.START_S2P = 32
                self.END_S2P = 40
                self.START_DS2P = 41
                self.END_DS2P = 48
                self.START_QA = 50
                self.END_QA = 58
                self.START_DQA = 59
                self.END_DQA = 66
                self.START_Q2B = 67
                self.END_Q2B = 76
                self.START_DQ2B = 77
                self.END_DQ2B = 84
                self.START_QEP = 85
                self.END_QEP = 94
                self.START_DQEP = 95
                self.END_DQEP = 102
                self.START_QBN = 103
                self.END_QBN = 112
                self.START_DQBN = 113
                self.END_DQBN = 125

        self.column_limits = [
            (self.START_R1_A, self.END_R1_A),
            (self.START_R1_Z, self.END_R1_Z),
            (self.START_S2N, self.END_S2N),
            (self.START_DS2N, self.END_DS2N),
            (self.START_S2P, self.END_S2P),
            (self.START_DS2P, self.END_DS2P),
            (self.START_QA, self.END_QA),
            (self.START_DQA, self.END_DQA),
            (self.START_Q2B, self.END_Q2B),
            (self.START_DQ2B, self.END_DQ2B),
            (self.START_QEP, self.END_QEP),
            (self.START_DQEP, self.END_DQEP),
            (self.START_QBN, self.END_QBN),
            (self.START_DQBN, self.END_DQBN),
        ]



from nuclearmasses.utils.element_converter import ElementConverter


class NUBASEFile(ElementConverter):
    """Easy access to where variables are in the NUBASE file.

    The NUBASE data file is formatted by location in the line, values exist
    between 2 specific columns in the line. Store the start and end locations
    in this class to allow simple access and stop the NUBASE parser having
    magic numbers.
    """

    def __init__(self, year: int):
        """Setup the values that locate the variable."""
        super().__init__()
        match year:
            case 1995:
                self.HEADER = 0
                self.FOOTER = 0
                self.START_A = 0
                self.END_A = 3
                self.START_Z = 4
                self.END_Z = 7
                self.START_STATE = 7
                self.END_STATE = 8
                self.START_ME = 18
                self.END_ME = 29
                self.START_DME = 29
                self.END_DME = 38
                self.START_ISOMER = 39
                self.END_ISOMER = 46
                self.START_DISOMER = 48
                self.END_DISOMER = 56
                self.START_HALFLIFEVALUE = 60
                self.END_HALFLIFEVALUE = 68
                self.START_HALFLIFEUNIT = 69
                self.END_HALFLIFEUNIT = 71
                self.START_HALFLIFEERROR = 72
                self.END_HALFLIFEERROR = 77
                self.START_SPIN = 79
                self.END_SPIN = 93
                self.START_DECAYSTRING = 106
                self.END_DECAYSTRING = None
            case 2003:
                self.HEADER = 0
                self.FOOTER = 0
                self.START_A = 0
                self.END_A = 3
                self.START_Z = 4
                self.END_Z = 7
                self.START_STATE = 7
                self.END_STATE = 8
                self.START_ME = 18
                self.END_ME = 29
                self.START_DME = 29
                self.END_DME = 38
                self.START_ISOMER = 39
                self.END_ISOMER = 46
                self.START_DISOMER = 48
                self.END_DISOMER = 56
                self.START_HALFLIFEVALUE = 60
                self.END_HALFLIFEVALUE = 68
                self.START_HALFLIFEUNIT = 69
                self.END_HALFLIFEUNIT = 71
                self.START_HALFLIFEERROR = 72
                self.END_HALFLIFEERROR = 77
                self.START_SPIN = 79
                self.END_SPIN = 93
                self.START_DECAYSTRING = 106
                self.END_DECAYSTRING = None
            case 2020:
                self.HEADER = 25
                self.FOOTER = 0
                self.START_A = 0
                self.END_A = 3
                self.START_Z = 4
                self.END_Z = 7
                self.START_STATE = 7
                self.END_STATE = 8
                self.START_ME = 18
                self.END_ME = 31
                self.START_DME = 31
                self.END_DME = 42
                self.START_ISOMER = 43
                self.END_ISOMER = 53
                self.START_DISOMER = 54
                self.END_DISOMER = 64
                self.START_HALFLIFEVALUE = 69
                self.END_HALFLIFEVALUE = 77
                self.START_HALFLIFEUNIT = 78
                self.END_HALFLIFEUNIT = 80
                self.START_HALFLIFEERROR = 81
                self.END_HALFLIFEERROR = 87
                self.START_SPIN = 88
                self.END_SPIN = 101
                self.START_ENSDF = 102
                self.END_ENSDF = 103
                self.START_YEAR = 114
                self.END_YEAR = 118
                self.START_DECAYSTRING = 119
                self.END_DECAYSTRING = None
            case _:
                self.HEADER = 0
                self.FOOTER = 0
                self.START_A = 0
                self.END_A = 3
                self.START_Z = 4
                self.END_Z = 7
                self.START_STATE = 7
                self.END_STATE = 8
                self.START_ME = 18
                self.END_ME = 29
                self.START_DME = 29
                self.END_DME = 38
                self.START_ISOMER = 39
                self.END_ISOMER = 46
                self.START_DISOMER = 48
                self.END_DISOMER = 56
                self.START_HALFLIFEVALUE = 60
                self.END_HALFLIFEVALUE = 68
                self.START_HALFLIFEUNIT = 69
                self.END_HALFLIFEUNIT = 71
                self.START_HALFLIFEERROR = 72
                self.END_HALFLIFEERROR = 77
                self.START_SPIN = 79
                self.END_SPIN = 93
                self.START_YEAR = 105
                self.END_YEAR = 109
                self.START_DECAYSTRING = 110
                self.END_DECAYSTRING = None

        self.column_limits = [
            (self.START_A, self.END_A),
            (self.START_Z, self.END_Z),
            (self.START_STATE, self.END_STATE),
            (self.START_ME, self.END_ME),
            (self.START_DME, self.END_DME),
            (self.START_ISOMER, self.END_ISOMER),
            (self.START_DISOMER, self.END_DISOMER),
            (self.START_HALFLIFEVALUE, self.END_HALFLIFEVALUE),
            (self.START_HALFLIFEUNIT, self.END_HALFLIFEUNIT),
            (self.START_HALFLIFEERROR, self.END_HALFLIFEERROR),
            (self.START_SPIN, self.END_SPIN),
            (self.START_DECAYSTRING, self.END_DECAYSTRING),
        ]

        if year > 2003:
            self.column_limits.insert(-1, (self.START_YEAR, self.END_YEAR))



import astropy  # type: ignore[import-untyped]
import numpy as np
import pandas as pd


class ElementConverter:
    """A utility class for converting between symbol and Z value

    This class provides bidirectional lookup functionality via two dictionaries:
    one mapping Z to symbol and the other symbol to Z.

    TODO: Create accessors function that do some argument validation
    """

    def __init__(self) -> None:
        """Construct the symbol -> Z and Z -> symbol dictionaries."""
        # fmt: off
        # Formatter wants to put each item on it's own line, I don't
        self.z_to_symbol: dict[int, str] = {
            0: "n", 1: "H", 2: "He", 3: "Li", 4: "Be", 5: "B", 6: "C", 7: "N", 8: "O", 9: "F",
            10: "Ne", 11: "Na", 12: "Mg", 13: "Al", 14: "Si", 15: "P", 16: "S", 17: "Cl", 18: "Ar", 19: "K",
            20: "Ca", 21: "Sc", 22: "Ti", 23: "V", 24: "Cr", 25: "Mn", 26: "Fe", 27: "Co", 28: "Ni", 29: "Cu",
            30: "Zn", 31: "Ga", 32: "Ge", 33: "As", 34: "Se", 35: "Br", 36: "Kr", 37: "Rb", 38: "Sr", 39: "Y",
            40: "Zr", 41: "Nb", 42: "Mo", 43: "Tc", 44: "Ru", 45: "Rh", 46: "Pd", 47: "Ag", 48: "Cd", 49: "In",
            50: "Sn", 51: "Sb", 52: "Te", 53: "I", 54: "Xe", 55: "Cs", 56: "Ba", 57: "La", 58: "Ce", 59: "Pr",
            60: "Nd", 61: "Pm", 62: "Sm", 63: "Eu", 64: "Gd", 65: "Tb", 66: "Dy", 67: "Ho", 68: "Er", 69: "Tm",
            70: "Yb", 71: "Lu", 72: "Hf", 73: "Ta", 74: "W", 75: "Re", 76: "Os", 77: "Ir", 78: "Pt", 79: "Au",
            80: "Hg", 81: "Tl", 82: "Pb", 83: "Bi", 84: "Po", 85: "At", 86: "Rn", 87: "Fr", 88: "Ra", 89: "Ac",
            90: "Th", 91: "Pa", 92: "U", 93: "Np", 94: "Pu", 95: "Am", 96: "Cm", 97: "Bk", 98: "Cf", 99: "Es",
            100: "Fm", 101: "Md", 102: "No", 103: "Lr", 104: "Rf", 105: "Db", 106: "Sg", 107: "Bh", 108: "Hs", 109: "Mt", # noqa: E501
            110: "Ds", 111: "Rg", 112: "Cn", 113: "Ed", 114: "Fl", 115: "Ef", 116: "Lv", 117: "Ts", 118: "Og"
        }
        # fmt: on

        # Switch the keys and values of the z_to_symbol dictionary so a user can access the Z value using the symbol
        self.symbol_to_z: dict[str, int] = {val: key for key, val in self.z_to_symbol.items()}

    def unit_to_seconds(self, unit_str: str) -> float:
        """Convert a time unit to a scale factor in seconds.

        Returns np.nan for non-valid time units.

        e.g. "s" -> 1.0, "min" -> 60.0, "keV" -> np.nan

        """
        if pd.isna(unit_str):
            return np.nan

        # Remove white space and make lower case to be consistent
        # String conversion may be redundant, but let's be complete
        cleaned_unit = str(unit_str).strip().lower()
        if not cleaned_unit:
            return np.nan

        # Use 'silent' so failure is written into the variable, not thrown as an exception
        unit = astropy.units.Unit(cleaned_unit, parse_strict="silent")
        # Check for failures as mentioned above
        if isinstance(unit, astropy.units.UnrecognizedUnit):
            return np.nan

        # We now know the unit is physically valid, but not that it's time related, e.g. minutes rather than km
        # Astropy allows us to check that
        if unit.physical_type != "time":
            return np.nan

        return float(unit.to(astropy.units.s))



from nuclearmasses.utils.element_converter import ElementConverter


class AMEMassFile(ElementConverter):
    """Easy access to where the variables are in the AME mass file."""

    def __init__(self, year: int):
        super().__init__()
        match year:
            case 1983:
                self.HEADER = 35
                self.FOOTER = 0
                self.START_Z = 11
                self.END_Z = 14
                self.START_A = 16
                self.END_A = 19
                self.START_ME = 29
                self.END_ME = 39
                self.START_DME = 41
                self.END_DME = 48
                self.START_BE_PER_A = 49
                self.END_BE_PER_A = 59
                self.START_DBE_PER_A = 61
                self.END_DBE_PER_A = 68
                self.START_BETA_DECAY_ENERGY = 76
                self.END_BETA_DECAY_ENERGY = 85
                self.START_DBETA_DECAY_ENERGY = 87
                self.END_DBETA_DECAY_ENERGY = 94
                self.START_AM = 97
                self.END_AM = 99
                self.START_MICRO_U = 100
                self.END_MICRO_U = 110
                self.START_MICRO_DU = 113
                self.END_MICRO_DU = 120
            case 1993:
                self.HEADER = 40
                self.FOOTER = 0
                self.START_Z = 11
                self.END_Z = 14
                self.START_A = 16
                self.END_A = 19
                self.START_ME = 29
                self.END_ME = 39
                self.START_DME = 41
                self.END_DME = 48
                self.START_BE_PER_A = 49
                self.END_BE_PER_A = 59
                self.START_DBE_PER_A = 61
                self.END_DBE_PER_A = 68
                self.START_BETA_DECAY_ENERGY = 76
                self.END_BETA_DECAY_ENERGY = 85
                self.START_DBETA_DECAY_ENERGY = 87
                self.END_DBETA_DECAY_ENERGY = 94
                self.START_AM = 97
                self.END_AM = 99
                self.START_MICRO_U = 100
                self.END_MICRO_U = 110
                self.START_MICRO_DU = 112
                self.END_MICRO_DU = 120
            case 1995:
                self.HEADER = 39
                self.FOOTER = 0
                self.START_Z = 11
                self.END_Z = 14
                self.START_A = 16
                self.END_A = 19
                self.START_ME = 29
                self.END_ME = 39
                self.START_DME = 41
                self.END_DME = 48
                self.START_BE_PER_A = 49
                self.END_BE_PER_A = 59
                self.START_DBE_PER_A = 61
                self.END_DBE_PER_A = 68
                self.START_BETA_DECAY_ENERGY = 76
                self.END_BETA_DECAY_ENERGY = 85
                self.START_DBETA_DECAY_ENERGY = 87
                self.END_DBETA_DECAY_ENERGY = 94
                self.START_AM = 97
                self.END_AM = 99
                self.START_MICRO_U = 100
                self.END_MICRO_U = 110
                self.START_MICRO_DU = 112
                self.END_MICRO_DU = 120
            case 2020:
                self.HEADER = 36
                self.FOOTER = 0
                self.START_Z = 11
                self.END_Z = 14
                self.START_A = 16
                self.END_A = 19
                self.START_ME = 29
                self.END_ME = 42
                self.START_DME = 43
                self.END_DME = 53
                self.START_BE_PER_A = 56
                self.END_BE_PER_A = 66
                self.START_DBE_PER_A = 69
                self.END_DBE_PER_A = 77
                self.START_BETA_DECAY_ENERGY = 82
                self.END_BETA_DECAY_ENERGY = 93
                self.START_DBETA_DECAY_ENERGY = 95
                self.END_DBETA_DECAY_ENERGY = 104
                self.START_AM = 106
                self.END_AM = 109
                self.START_MICRO_U = 110
                self.END_MICRO_U = 120
                self.START_MICRO_DU = 124
                self.END_MICRO_DU = 135
            case _:
                self.HEADER = 39
                self.FOOTER = 0
                self.START_Z = 11
                self.END_Z = 14
                self.START_A = 16
                self.END_A = 19
                self.START_ME = 29
                self.END_ME = 41
                self.START_DME = 42
                self.END_DME = 53
                self.START_BE_PER_A = 54
                self.END_BE_PER_A = 64
                self.START_DBE_PER_A = 65
                self.END_DBE_PER_A = 72
                self.START_BETA_DECAY_ENERGY = 76
                self.END_BETA_DECAY_ENERGY = 86
                self.START_DBETA_DECAY_ENERGY = 87
                self.END_DBETA_DECAY_ENERGY = 95
                self.START_AM = 96
                self.END_AM = 99
                self.START_MICRO_U = 100
                self.END_MICRO_U = 112
                self.START_MICRO_DU = 113
                self.END_MICRO_DU = 120

        self.column_limits = [
            (self.START_Z, self.END_Z),
            (self.START_A, self.END_A),
            (self.START_ME, self.END_ME),
            (self.START_DME, self.END_DME),
            (self.START_BE_PER_A, self.END_BE_PER_A),
            (self.START_DBE_PER_A, self.END_DBE_PER_A),
            (self.START_BETA_DECAY_ENERGY, self.END_BETA_DECAY_ENERGY),
            (self.START_DBETA_DECAY_ENERGY, self.END_DBETA_DECAY_ENERGY),
            (self.START_AM, self.END_AM),
            (self.START_MICRO_U, self.END_MICRO_U),
            (self.START_MICRO_DU, self.END_MICRO_DU),
        ]



from nuclearmasses.mass_table import MassTable


def test_get_nubase_datafile():
    mt = MassTable()

    year = 1995
    assert mt._get_nubase_datafile(year) == mt.data_path / str(year) / "nubtab97.asc"
    year = 2003
    assert mt._get_nubase_datafile(year) == mt.data_path / str(year) / "nubtab03.asc"
    year = 2012
    assert mt._get_nubase_datafile(year) == mt.data_path / str(year) / "nubtab12.asc"
    year = 2016
    assert mt._get_nubase_datafile(year) == mt.data_path / str(year) / "nubase2016.txt"
    year = 2020
    assert mt._get_nubase_datafile(year) == mt.data_path / str(year) / "nubase_1.mas20"


def test_get_ame_datafiles():
    mt = MassTable()

    year = 1983
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(1983)
    assert mass == data_path / "mass.mas83"
    assert reaction01 == data_path / "rct1.mas83"
    assert reaction02 == data_path / "rct2.mas83"

    year = 1993
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(1993)
    assert mass == data_path / "mass_exp.mas93"
    assert reaction01 == data_path / "rct1_exp.mas93"
    assert reaction02 == data_path / "rct2_exp.mas93"

    year = 1995
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(1995)
    assert mass == data_path / "mass_exp.mas95"
    assert reaction01 == data_path / "rct1_exp.mas95"
    assert reaction02 == data_path / "rct2_exp.mas95"

    year = 2012
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(2012)
    assert mass == data_path / "mass.mas12"
    assert reaction01 == data_path / "rct1.mas12"
    assert reaction02 == data_path / "rct2.mas12"

    year = 2016
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(2016)
    assert mass == data_path / "mass16.txt"
    assert reaction01 == data_path / "rct1-16.txt"
    assert reaction02 == data_path / "rct2-16.txt"

    year = 2020
    data_path = mt.data_path / str(year)
    mass, reaction01, reaction02 = mt._get_ame_datafiles(2020)
    assert mass == data_path / "mass.mas20"
    assert reaction01 == data_path / "rct1.mas20"
    assert reaction02 == data_path / "rct2.mas20"



import logging
import pathlib
import typing

import pandas as pd

from nuclearmasses.io.nubase_file import NUBASEFile


class NUBASEParser(NUBASEFile):
    """Parse the NUBASE data file.

    A collection of functions to parse the weird format of the NUBASE file.
    """

    def __init__(self, filename: pathlib.Path, year: int):
        """Set the file to read and the table year."""
        super().__init__(year)
        self.filename: pathlib.Path = filename
        self.year: int = year
        self.unit_replacements: dict[str, str] = {
            r"y$": "yr",
            r"^m$": "min",
        }
        logging.info(f"Reading {self.filename} from {self.year}")

    def _column_names(self) -> list[str]:
        """Set the column name depending on the year"""
        match self.year:
            case 1995 | 2003:
                return [
                    "A",
                    "Z",
                    "State",
                    "NUBASEMassExcess",
                    "NUBASEMassExcessError",
                    "IsomerEnergy",
                    "IsomerEnergyError",
                    "HalfLifeValue",
                    "HalfLifeUnit",
                    "HalfLifeError",
                    "Spin",
                    "DecayModes",
                ]
            case _:
                return [
                    "A",
                    "Z",
                    "State",
                    "NUBASEMassExcess",
                    "NUBASEMassExcessError",
                    "IsomerEnergy",
                    "IsomerEnergyError",
                    "HalfLifeValue",
                    "HalfLifeUnit",
                    "HalfLifeError",
                    "Spin",
                    "DiscoveryYear",
                    "DecayModes",
                ]

    def _data_types(self) -> dict:
        """Set the data type depending on the year"""
        match self.year:
            case 1995 | 2003:
                return {
                    "Symbol": "string",
                    "A": "Int64",
                    "Z": "Int64",
                    "N": "Int64",
                    "Experimental": "boolean",
                    # "State": "Int64",
                    "NUBASEMassExcess": "float64",
                    "NUBASEMassExcessError": "float64",
                    # "IsomerEnergy": "float64",
                    # "IsomerEnergyError": "float64",
                    "HalfLifeValue": "float64",
                    "HalfLifeUnit": "string",
                    "HalfLifeError": "float64",
                    "HalfLifeSeconds": "float64",
                    "HalfLifeErrorSeconds": "float64",
                    "Spin": "string",
                    "DecayModes": "string",
                }
            case _:
                return {
                    "Symbol": "string",
                    "A": "Int64",
                    "Z": "Int64",
                    "N": "Int64",
                    "Experimental": "boolean",
                    # "State": "Int64",
                    "NUBASEMassExcess": "float64",
                    "NUBASEMassExcessError": "float64",
                    # "IsomerEnergy": "float64",
                    # "IsomerEnergyError": "float64",
                    "HalfLifeValue": "float64",
                    "HalfLifeUnit": "string",
                    "HalfLifeError": "float64",
                    "HalfLifeSeconds": "float64",
                    "HalfLifeErrorSeconds": "float64",
                    "Spin": "string",
                    "DiscoveryYear": "Int64",
                    "DecayModes": "string",
                }

    def _na_values(self) -> dict:
        """Set the columns that have placeholder values"""
        match self.year:
            case 1995:
                return {
                    "NUBASEMassExcess": [""],
                    "NUBASEMassExcessError": [""],
                    "State": [""],
                    "HalfLifeValue": [""],
                    "HalfLifeUnit": [""],
                    "HalfLifeError": [""],
                    "Spin": [""],
                    "DecayModes": [""],
                }
            case _:
                return {
                    "State": [""],
                    "NUBASEMassExcess": [""],
                    "NUBASEMassExcessError": [""],
                    "HalfLifeValue": ["", "p-unst", "p-unst#"],
                    "HalfLifeUnit": [""],
                    "HalfLifeError": [""],
                    "DiscoveryYear": [""],
                    "DecayModes": [""],
                }

    def read_file(self) -> pd.DataFrame:
        """Read the file using it's known format

        The AMEMassFile and other functions in this class have hopefully sanitized the
        column names, data types and locations of the date so we can not make the generic
        call to parse the file.
        """
        try:
            df = pd.read_fwf(
                self.filename,
                colspecs=typing.cast(typing.Sequence[tuple[int, int]], self.column_limits),  # appease mypy
                names=self._column_names(),
                na_values=self._na_values(),
                keep_default_na=False,
                on_bad_lines="warn",
                skiprows=self.HEADER,
                skipfooter=self.FOOTER,
            )
            # We use the NUBASE data to define whether or not an isotope is experimentally measured,
            df["Experimental"] = ~df["NUBASEMassExcess"].astype("string").str.contains("#", na=False)
            # Once we have used the '#' to determine if it's experimental or not, we can remove all instances of it
            df.replace("#", "", regex=True, inplace=True)

            df["TableYear"] = self.year
            df["N"] = pd.to_numeric(df["A"]) - pd.to_numeric(df["Z"])
            df["Symbol"] = pd.to_numeric(df["Z"]).map(self.z_to_symbol)
            # For the moment, we will ignore anything this is not the ground state
            df = df[df["State"] == 0]
            # As 'State' is now necessarily 0 and the Isomer columns are empty, drop them.
            df = df.drop(columns=["State", "IsomerEnergy", "IsomerEnergyError"])

            # Convert stable isotopes into ones with enormous lifetimes with zero error so we can cast
            # pandas v3 became much stricter with type conversions so convert to object (from string) so
            # we can assign a float without breaking other parts of the code
            df["HalfLifeValue"] = df["HalfLifeValue"].astype("object")
            df["HalfLifeError"] = df["HalfLifeError"].astype("object")

            mask = df["HalfLifeValue"] == "stbl"
            df.loc[mask, ["HalfLifeValue", "HalfLifeUnit", "HalfLifeError"]] = (99.99, "Zyr", 0.0)

            df["HalfLifeValue"] = df["HalfLifeValue"].astype("string").str.replace(r"[<>?~]", "", regex=True)
            # We'll be lazy here and remove any characters in this column. Future us will parse this properly
            df["HalfLifeError"] = df["HalfLifeError"].astype("string").str.replace(r"[<>?~a-z]", "", regex=True)

            # Use the 3 half-life columns to create 2 new columns with units of seconds
            df["HalfLifeUnit"] = df["HalfLifeUnit"].astype("string")
            for pattern, replacement in self.unit_replacements.items():
                df["HalfLifeUnit"] = df["HalfLifeUnit"].str.replace(pattern, replacement, regex=True)

            # Ensure numeric values
            for col in ["HalfLifeValue", "HalfLifeError"]:
                df[col] = pd.to_numeric(df[col], errors="coerce")
            # Pre-compute unit -> second conversions
            unit_map = df["HalfLifeUnit"].map(self.unit_to_seconds)

            df["HalfLifeSeconds"] = df["HalfLifeValue"] * unit_map
            df["HalfLifeErrorSeconds"] = df["HalfLifeError"] * unit_map
        except ValueError as e:
            print(f"Parsing error: {e}")

        return df.astype(self._data_types())



import logging
import pathlib

import pandas as pd

from nuclearmasses.io.ame_reaction_2_file import AMEReactionFileTwo


class AMEReactionParserTwo(AMEReactionFileTwo):
    """Parse the second AME reaction file.

    The format is known but I don't think python can easily parse it.
    """

    def __init__(self, filename: pathlib.Path, year: int):
        """Set the file to read and table year."""
        self.filename = filename
        self.year = year
        super().__init__(self.year)
        logging.info(f"Reading {self.filename} from {self.year}")

    def _column_names(self) -> list[str]:
        """Set the column name depending on the year"""
        match self.year:
            case _:
                return [
                    "A",
                    "Z",
                    "OneNeutronSeparationEnergy",
                    "OneNeutronSeparationEnergyError",
                    "OneProtonSeparationEnergy",
                    "OneProtonSeparationEnergyError",
                    "QFourBeta",
                    "QFourBetaError",
                    "QDeuteronAlpha",
                    "QDeuteronAlphaError",
                    "QProtonAlpha",
                    "QProtonAlphaError",
                    "QNeutronAlpha",
                    "QNeutronAlphaError",
                ]

    def _data_types(self) -> dict:
        """Set the data type depending on the year"""
        match self.year:
            case _:
                return {
                    "TableYear": "Int64",
                    "Symbol": "string",
                    "A": "Int64",
                    "Z": "Int64",
                    "N": "Int64",
                    "OneNeutronSeparationEnergy": "float64",
                    "OneNeutronSeparationEnergyError": "float64",
                    "OneProtonSeparationEnergy": "float64",
                    "OneProtonSeparationEnergyError": "float64",
                    "QFourBeta": "float64",
                    "QFourBetaError": "float64",
                    "QDeuteronAlpha": "float64",
                    "QDeuteronAlphaError": "float64",
                    "QProtonAlpha": "float64",
                    "QProtonAlphaError": "float64",
                    "QNeutronAlpha": "float64",
                    "QNeutronAlphaError": "float64",
                }

    def _na_values(self) -> dict:
        """Set the columns that have placeholder values"""
        match self.year:
            case _:
                return {
                    "A": [""],
                    "OneNeutronSeparationEnergy": ["", "*"],
                    "OneNeutronSeparationEnergyError": ["", "*"],
                    "OneProtonSeparationEnergy": ["", "*"],
                    "OneProtonSeparationEnergyError": ["", "*"],
                    "QFourBeta": ["", "*"],
                    "QFourBetaError": ["", "*"],
                    "QDeuteronAlpha": ["", "*"],
                    "QDeuteronAlphaError": ["", "*"],
                    "QProtonAlpha": ["", "*"],
                    "QProtonAlphaError": ["", "*"],
                    "QNeutronAlpha": ["", "*"],
                    "QNeutronAlphaError": ["", "*"],
                }

    def read_file(self) -> pd.DataFrame:
        """Read the file using it's known format

        The AMEReactionFileTwo and other functions in this class have hopefully sanitized the
        column names, data types and locations of the date so we can not make the generic
        call to parse the file.
        """
        try:
            df = pd.read_fwf(
                self.filename,
                colspecs=self.column_limits,
                names=self._column_names(),
                na_values=self._na_values(),
                keep_default_na=False,
                on_bad_lines="warn",
                skiprows=self.HEADER,
                skipfooter=self.FOOTER,
            )
            # We use the NUBASE data to define whether or not an isotope is experimentally measured,
            # so for this data we'll just drop any and all '#' characters
            df.replace("#", "", regex=True, inplace=True)

            if self.year == 1983:
                # The column headers and units are repeated in the 1983 table
                df = df[(df["A"] != "A") & (df["Z"] != "")]
                # The A value is not in the column if it doesn't change so we need to fill down
                df["A"] = df["A"].ffill()
            elif self.year == 2020:
                # The column headers and units are repeated in the 2020 table
                df = df[(df["A"] != "A") & (df["Z"] != "Z")]

            # Repeated column heading also means we have to cast to create new columns
            df["TableYear"] = self.year
            df["N"] = pd.to_numeric(df["A"]) - pd.to_numeric(df["Z"])
            df["Symbol"] = pd.to_numeric(df["Z"]).map(self.z_to_symbol)
        except ValueError as e:
            print(f"Parsing error: {e}")

        return df.astype(self._data_types())



import io

import pandas as pd
import pandas.testing as pdt

from nuclearmasses.io.ame_reaction_2_parse import AMEReactionParserTwo


def test_1983_rct2():
    line = io.StringIO(
        " 204 Tl  81    7853      17      5702.8     1.7  -13480     120     12613.7     1.8    8608.4     1.8    7180      50"
    )
    parser = AMEReactionParserTwo(line, 1983)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1983],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [7853],
            "OneNeutronSeparationEnergyError": [17],
            "OneProtonSeparationEnergy": [5702.8],
            "OneProtonSeparationEnergyError": [1.7],
            "QFourBeta": [-13480],
            "QFourBetaError": [120],
            "QDeuteronAlpha": [12613.7],
            "QDeuteronAlphaError": [1.8],
            "QProtonAlpha": [8608.4],
            "QProtonAlphaError": [1.8],
            "QNeutronAlpha": [7180],
            "QNeutronAlphaError": [50],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1993_rct2():
    line = io.StringIO(
        " 204 Tl  81    6655.82    0.29   6365.32    1.26 -12492.85   71.39  13712.87    1.23   8183.14    1.25   7690.59   15.05"
    )
    parser = AMEReactionParserTwo(line, 1993)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1993],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6655.86],
            "OneNeutronSeparationEnergyError": [0.29],
            "OneProtonSeparationEnergy": [6365.35],
            "OneProtonSeparationEnergyError": [1.26],
            "QFourBeta": [-12492.85],
            "QFourBetaError": [71.39],
            "QDeuteronAlpha": [13712.87],
            "QDeuteronAlphaError": [1.23],
            "QProtonAlpha": [8183.14],
            "QProtonAlphaError": [1.25],
            "QNeutronAlpha": [7690.59],
            "QNeutronAlphaError": [15.05],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1995_rct2():
    line = io.StringIO(
        " 204 Tl  81    6655.86    0.29   6365.35    1.26 -12494.05   92.85  13713.05    1.23   8183.32    1.24   7702.97    3.35"
    )
    parser = AMEReactionParserTwo(line, 1995)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1995],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6655.86],
            "OneNeutronSeparationEnergyError": [0.29],
            "OneProtonSeparationEnergy": [6365.35],
            "OneProtonSeparationEnergyError": [1.26],
            "QFourBeta": [-12494.05],
            "QFourBetaError": [92.85],
            "QDeuteronAlpha": [13713.05],
            "QDeuteronAlphaError": [1.23],
            "QProtonAlpha": [8183.32],
            "QProtonAlphaError": [1.24],
            "QNeutronAlpha": [7702.97],
            "QNeutronAlphaError": [3.35],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2003_rct2():
    line = io.StringIO(
        " 204 Tl  81    6656.10    0.29   6365.82    1.25 -12470.66   24.01  13710.69    1.15   8181.34    1.16   7701.54    3.34"
    )
    parser = AMEReactionParserTwo(line, 2003)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2003],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6656.10],
            "OneNeutronSeparationEnergyError": [0.29],
            "OneProtonSeparationEnergy": [6365.82],
            "OneProtonSeparationEnergyError": [1.25],
            "QFourBeta": [-12470.66],
            "QFourBetaError": [24.01],
            "QDeuteronAlpha": [13710.69],
            "QDeuteronAlphaError": [1.15],
            "QProtonAlpha": [8181.34],
            "QProtonAlphaError": [1.16],
            "QNeutronAlpha": [7701.54],
            "QNeutronAlphaError": [3.34],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2012_rct2():
    line = io.StringIO(
        " 204 Tl  81    6656.09    0.29   6365.80    1.25 -12470.19   22.31  13710.68    1.14   8181.16    1.15   7701.67    3.33"
    )
    parser = AMEReactionParserTwo(line, 2012)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2012],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6656.09],
            "OneNeutronSeparationEnergyError": [0.29],
            "OneProtonSeparationEnergy": [6365.80],
            "OneProtonSeparationEnergyError": [1.25],
            "QFourBeta": [-12470.19],
            "QFourBetaError": [22.31],
            "QDeuteronAlpha": [13710.68],
            "QDeuteronAlphaError": [1.14],
            "QProtonAlpha": [8181.16],
            "QProtonAlphaError": [1.15],
            "QNeutronAlpha": [7701.67],
            "QNeutronAlphaError": [3.33],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2016_rct2():
    line = io.StringIO(
        " 204 Tl  81    6656.08    0.29   6365.85    1.25 -12470.71   22.32  13709.99    1.06   8180.45    1.07   7700.97    3.31"
    )
    parser = AMEReactionParserTwo(line, 2016)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2016],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6656.08],
            "OneNeutronSeparationEnergyError": [0.29],
            "OneProtonSeparationEnergy": [6365.85],
            "OneProtonSeparationEnergyError": [1.25],
            "QFourBeta": [-12470.71],
            "QFourBetaError": [22.32],
            "QDeuteronAlpha": [13709.99],
            "QDeuteronAlphaError": [1.06],
            "QProtonAlpha": [8180.45],
            "QProtonAlphaError": [1.07],
            "QNeutronAlpha": [7700.97],
            "QNeutronAlphaError": [3.31],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2020_rct2():
    line = io.StringIO(
        " 204 Tl  81    6656.0787    0.2907   6365.8379    1.2542 -12470.8182   22.6974  13710.0469    1.0612   8180.5147    1.0721   7701.0380    3.3084"
    )
    parser = AMEReactionParserTwo(line, 2020)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2020],
            "Symbol": ["Tl"],
            "A": [204],
            "Z": [81],
            "N": [123],
            "OneNeutronSeparationEnergy": [6656.0787],
            "OneNeutronSeparationEnergyError": [0.2907],
            "OneProtonSeparationEnergy": [6365.8379],
            "OneProtonSeparationEnergyError": [1.2542],
            "QFourBeta": [-12470.8182],
            "QFourBetaError": [22.6974],
            "QDeuteronAlpha": [13710.0469],
            "QDeuteronAlphaError": [1.0612],
            "QProtonAlpha": [8180.5147],
            "QProtonAlphaError": [1.0721],
            "QNeutronAlpha": [7701.0380],
            "QNeutronAlphaError": [3.3084],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)



import io

import pandas as pd
import pandas.testing as pdt

from nuclearmasses.io.ame_mass_parse import AMEMassParser


def test_1983_mass():
    # 67Fe was randomly selected to test the AME mass parsing. According to AME it didn't exist in 1983 so will use 67Ni
    line = io.StringIO(
        "0 11   39   28   67 Ni +n2p  -63742.471   19.056 582618.683   19.066    B-   3560.871   20.646   66 931570.167   20.457  -.0"
    )
    parser = AMEMassParser(line, 1983)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1983],
            "Symbol": ["Ni"],
            "A": [67],
            "Z": [28],
            "N": [39],
            "AMEMassExcess": [-63742.471],
            "AMEMassExcessError": [19.056],
            "BindingEnergyPerA": [582618.683 / 67],
            "BindingEnergyPerAError": [19.066 / 67],
            "BetaDecayEnergy": [3560.871],
            "BetaDecayEnergyError": [20.646],
            "AtomicMass": [66.931579167],
            "AtomicMassError": [20.457 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1993_mass():
    line = io.StringIO(
        "0 15   41   26   67 Fe    x  -46574.693  465.747 567012.139  465.747    B-   8746.727  543.150   66 950000.000  500.000"
    )
    parser = AMEMassParser(line, 1993)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1993],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-46574.693],
            "AMEMassExcessError": [465.747],
            "BindingEnergyPerA": [567012.133 / 67],
            "BindingEnergyPerAError": [465.747 / 67],
            "BetaDecayEnergy": [8746.727],
            "BetaDecayEnergyError": [543.150],
            "AtomicMass": [66.950000000],
            "AtomicMassError": [500.0 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_1995_mass():
    # Yes the 1995 line is identical to the 1993 line
    line = io.StringIO(
        "  15   41   26   67 Fe    x  -46574.693  465.747 567012.133  465.747    B-   8746.727  543.150   66 950000.000  500.000"
    )
    parser = AMEMassParser(line, 1995)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1995],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-46574.693],
            "AMEMassExcessError": [465.747],
            "BindingEnergyPerA": [567012.133 / 67],
            "BindingEnergyPerAError": [465.747 / 67],
            "BetaDecayEnergy": [8746.727],
            "BetaDecayEnergyError": [543.150],
            "AtomicMass": [66.950000000],
            "AtomicMassError": [500.0 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2003_mass():
    line = io.StringIO(
        "  15   41   26   67 Fe    x  -45692.348    415.570     8449.695    6.203 B-   9368.702  523.438  66 950947.244    446.132"
    )
    parser = AMEMassParser(line, 2003)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2003],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-45692.348],
            "AMEMassExcessError": [415.570],
            "BindingEnergyPerA": [8449.695],
            "BindingEnergyPerAError": [6.203],
            "BetaDecayEnergy": [9368.702],
            "BetaDecayEnergyError": [523.438],
            "AtomicMass": [66.950947244],
            "AtomicMassError": [446.132 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2012_mass():
    line = io.StringIO(
        "  15   41   26   67 Fe    x  -46068.530    217.972     8455.310    3.253 B-   9253.245  218.067  66 950543.395    234.002"
    )
    parser = AMEMassParser(line, 2012)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2012],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-46068.530],
            "AMEMassExcessError": [217.972],
            "BindingEnergyPerA": [8455.310],
            "BindingEnergyPerAError": [3.253],
            "BetaDecayEnergy": [9253.245],
            "BetaDecayEnergyError": [218.067],
            "AtomicMass": [66.950543395],
            "AtomicMassError": [234.002 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2016_mass():
    line = io.StringIO(
        "  15   41   26   67 Fe    x  -45610.155    270.285     8448.469    4.034 B-   9711.620  270.362  66 951035.482    290.163"
    )
    parser = AMEMassParser(line, 2016)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2016],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-45610.155],
            "AMEMassExcessError": [270.285],
            "BindingEnergyPerA": [8448.469],
            "BindingEnergyPerAError": [4.034],
            "BetaDecayEnergy": [9711.620],
            "BetaDecayEnergyError": [270.362],
            "AtomicMass": [66.951035482],
            "AtomicMassError": [290.163 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2020_mass():
    line = io.StringIO(
        "  15   41   26   67 Fe    x  -45708.416       3.819      8449.9359     0.0570  B-   9613.3678     7.4900   66 950930.000       4.100"
    )
    parser = AMEMassParser(line, 2020)
    parser.HEADER = 0
    parser.FOOTER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2020],
            "Symbol": ["Fe"],
            "A": [67],
            "Z": [26],
            "N": [41],
            "AMEMassExcess": [-45708.416],
            "AMEMassExcessError": [3.819],
            "BindingEnergyPerA": [8449.9359],
            "BindingEnergyPerAError": [0.0570],
            "BetaDecayEnergy": [9613.3678],
            "BetaDecayEnergyError": [7.4900],
            "AtomicMass": [66.950930],
            "AtomicMassError": [4.100 / 1.0e6],
        }
    )
    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)



import io

import pandas as pd
import pandas.testing as pdt

from nuclearmasses.io.nubase_parse import NUBASEParser


def test_1995_nubase():
    line = io.StringIO(
        "168 0670   168Ho  -60085       29                              2.99   m 0.07   3+            94           B-=100"
    )
    parser = NUBASEParser(line, 1995)
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [1995],
            "Experimental": [True],
            "Symbol": ["Ho"],
            "A": [168],
            "Z": [67],
            "N": [101],
            "NUBASEMassExcess": [-60085],
            "NUBASEMassExcessError": [29],
            "HalfLifeValue": [2.99],
            "HalfLifeUnit": ["min"],
            "HalfLifeError": [0.07],
            "HalfLifeSeconds": [179.4],
            "HalfLifeErrorSeconds": [4.2],
            "Spin": ["3+"],
            "DecayModes": ["B-=100"],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2003_nubase():
    line = io.StringIO(
        "168 0670   168Ho  -60070       30                              2.99   m 0.07   3+            94           B-=100"
    )
    parser = NUBASEParser(line, 2003)
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2003],
            "Experimental": [True],
            "Symbol": ["Ho"],
            "A": [168],
            "Z": [67],
            "N": [101],
            "NUBASEMassExcess": [-60070],
            "NUBASEMassExcessError": [30],
            "HalfLifeValue": [2.99],
            "HalfLifeUnit": ["min"],
            "HalfLifeError": [0.07],
            "HalfLifeSeconds": [179.4],
            "HalfLifeErrorSeconds": [4.2],
            "Spin": ["3+"],
            "DecayModes": ["B-=100"],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2012_nubase():
    line = io.StringIO(
        "168 0670   168Ho  -60060       30                              2.99   m 0.07   3+            10          1960 B-=100"
    )
    parser = NUBASEParser(line, 2012)
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2012],
            "Experimental": [True],
            "Symbol": ["Ho"],
            "A": [168],
            "Z": [67],
            "N": [101],
            "NUBASEMassExcess": [-60060],
            "NUBASEMassExcessError": [30],
            "HalfLifeValue": [2.99],
            "HalfLifeUnit": ["min"],
            "HalfLifeError": [0.07],
            "HalfLifeSeconds": [179.4],
            "HalfLifeErrorSeconds": [4.2],
            "Spin": ["3+"],
            "DiscoveryYear": [1960],
            "DecayModes": ["B-=100"],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2016_nubase():
    line = io.StringIO(
        "168 0670   168Ho  -60060       30                              2.99   m 0.07   3+            10          1960 B-=100"
    )
    parser = NUBASEParser(line, 2016)
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2016],
            "Experimental": [True],
            "Symbol": ["Ho"],
            "A": [168],
            "Z": [67],
            "N": [101],
            "NUBASEMassExcess": [-60060],
            "NUBASEMassExcessError": [30],
            "HalfLifeValue": [2.99],
            "HalfLifeUnit": ["min"],
            "HalfLifeError": [0.07],
            "HalfLifeSeconds": [179.4],
            "HalfLifeErrorSeconds": [4.2],
            "Spin": ["3+"],
            "DiscoveryYear": [1960],
            "DecayModes": ["B-=100"],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)


def test_2020_nubase():
    line = io.StringIO(
        "168 0670   168Ho  -60060         30                                     2.99   m 0.07   3+            10          1960 B-=100"
    )
    parser = NUBASEParser(line, 2020)
    parser.HEADER = 0
    df = parser.read_file()

    expected = pd.DataFrame(
        {
            "TableYear": [2020],
            "Experimental": [True],
            "Symbol": ["Ho"],
            "A": [168],
            "Z": [67],
            "N": [101],
            "NUBASEMassExcess": [-60060],
            "NUBASEMassExcessError": [30],
            "HalfLifeValue": [2.99],
            "HalfLifeUnit": ["min"],
            "HalfLifeError": [0.07],
            "HalfLifeSeconds": [179.4],
            "HalfLifeErrorSeconds": [4.2],
            "Spin": ["3+"],
            "DiscoveryYear": [1960],
            "DecayModes": ["B-=100"],
        }
    )

    expected = expected.astype(parser._data_types())

    pdt.assert_frame_equal(df, expected, check_like=True)



from nuclearmasses.utils.element_converter import ElementConverter


class AMEReactionFileTwo(ElementConverter):
    """Easy access to where the variables are in the second AME reaction file."""

    def __init__(self, year: int):
        super().__init__()
        match year:
            case 1983:
                self.HEADER = 30
                self.FOOTER = 0
                self.START_R2_A = 1
                self.END_R2_A = 4
                self.START_R2_Z = 8
                self.END_R2_Z = 11
                self.START_SN = 14
                self.END_SN = 22
                self.START_DSN = 24
                self.END_DSN = 28
                self.START_SP = 30
                self.END_SP = 40
                self.START_DSP = 42
                self.END_DSP = 48
                self.START_Q4B = 49
                self.END_Q4B = 57
                self.START_DQ4B = 60
                self.END_DQ4B = 65
                self.START_QDA = 68
                self.END_QDA = 76
                self.START_DQDA = 78
                self.END_DQDA = 84
                self.START_QPA = 86
                self.END_QPA = 94
                self.START_DQPA = 96
                self.END_DQPA = 102
                self.START_QNA = 103
                self.END_QNA = 112
                self.START_DQNA = 114
                self.END_DQNA = 120
            case 2020:
                self.HEADER = 37
                self.FOOTER = 15
                self.START_R2_A = 1
                self.END_R2_A = 4
                self.START_R2_Z = 8
                self.END_R2_Z = 11
                self.START_SN = 14
                self.END_SN = 24
                self.START_DSN = 25
                self.END_DSN = 34
                self.START_SP = 36
                self.END_SP = 46
                self.START_DSP = 47
                self.END_DSP = 56
                self.START_Q4B = 57
                self.END_Q4B = 68
                self.START_DQ4B = 69
                self.END_DQ4B = 78
                self.START_QDA = 79
                self.END_QDA = 90
                self.START_DQDA = 91
                self.END_DQDA = 100
                self.START_QPA = 101
                self.END_QPA = 112
                self.START_DQPA = 113
                self.END_DQPA = 122
                self.START_QNA = 123
                self.END_QNA = 134
                self.START_DQNA = 135
                self.END_DQNA = 144
            case _:
                match year:
                    case 1995 | 2003 | 2012 | 2016:
                        self.HEADER = 39
                    case 1993:
                        self.HEADER = 40
                self.FOOTER = 0
                self.START_R2_A = 1
                self.END_R2_A = 4
                self.START_R2_Z = 8
                self.END_R2_Z = 11
                self.START_SN = 14
                self.END_SN = 22
                self.START_DSN = 23
                self.END_DSN = 30
                self.START_SP = 32
                self.END_SP = 40
                self.START_DSP = 41
                self.END_DSP = 48
                self.START_Q4B = 49
                self.END_Q4B = 58
                self.START_DQ4B = 59
                self.END_DQ4B = 66
                self.START_QDA = 67
                self.END_QDA = 76
                self.START_DQDA = 77
                self.END_DQDA = 84
                self.START_QPA = 85
                self.END_QPA = 94
                self.START_DQPA = 95
                self.END_DQPA = 102
                self.START_QNA = 103
                self.END_QNA = 112
                self.START_DQNA = 113
                self.END_DQNA = 125

        self.column_limits = [
            (self.START_R2_A, self.END_R2_A),
            (self.START_R2_Z, self.END_R2_Z),
            (self.START_SN, self.END_SN),
            (self.START_DSN, self.END_DSN),
            (self.START_SP, self.END_SP),
            (self.START_DSP, self.END_DSP),
            (self.START_Q4B, self.END_Q4B),
            (self.START_DQ4B, self.END_DQ4B),
            (self.START_QDA, self.END_QDA),
            (self.START_DQDA, self.END_DQDA),
            (self.START_QPA, self.END_QPA),
            (self.START_DQPA, self.END_DQPA),
            (self.START_QNA, self.END_QNA),
            (self.START_DQNA, self.END_DQNA),
        ]



from nuclearmasses.utils.element_converter import ElementConverter


def test_Z_to_symbol():
    converter = ElementConverter()

    assert converter.z_to_symbol[0] == "n"
    assert converter.z_to_symbol[6] == "C"
    assert converter.z_to_symbol[104] == "Rf"


def test_symbol_to_Z():
    converter = ElementConverter()

    assert converter.symbol_to_z["Al"] == 13
    assert converter.symbol_to_z["Fe"] == 26
    assert converter.symbol_to_z["Po"] == 84



import importlib.resources
import pathlib

import pandas as pd

from nuclearmasses.io.ame_mass_parse import AMEMassParser
from nuclearmasses.io.ame_reaction_1_parse import AMEReactionParserOne
from nuclearmasses.io.ame_reaction_2_parse import AMEReactionParserTwo
from nuclearmasses.io.nubase_parse import NUBASEParser


class MassTable:
    """Class for all of the mass data.

    Internally there are separate dataframes for the NUBASE and AME data as well as a combined one for all data
    """

    def __init__(self):
        """Do all of the work at construction."""
        self.data_path = importlib.resources.files("nuclearmasses.data")
        self.nubase_years = [1995, 2003, 2012, 2016, 2020]
        self.nubase = pd.concat([self._parse_nubase_data(y) for y in self.nubase_years], ignore_index=True)
        self.ame_years = [1983, 1993, 1995, 2003, 2012, 2016, 2020]
        self.ame = pd.concat([self._parse_ame_data(y) for y in self.ame_years], ignore_index=True)
        self.full_data = self._combine_all_data()
        self._do_indexing()

    def _get_nubase_datafile(self, year: int) -> pathlib.Path:
        """Use the given year to locate the NUBASE mass table file and return the absolute path."""
        nubase_mass = self.data_path / pathlib.Path(str(year))
        nubase_mass = nubase_mass.resolve()

        match year:
            case 1995:
                nubase_mass = nubase_mass / "nubtab97.asc"
            case 2003:
                nubase_mass = nubase_mass / "nubtab03.asc"
            case 2012:
                nubase_mass = nubase_mass / "nubtab12.asc"
            case 2016:
                nubase_mass = nubase_mass / "nubase2016.txt"
            case 2020:
                nubase_mass = nubase_mass / "nubase_1.mas20"

        return nubase_mass

    def _get_ame_datafiles(self, year: int) -> tuple[pathlib.Path, pathlib.Path, pathlib.Path]:
        """Use the given year to locate the 3 AME data file and return the absolute paths."""
        data_dir = self.data_path / pathlib.Path(str(year))
        data_dir = data_dir.resolve()

        match year:
            case 1983:
                ame_mass = data_dir / "mass.mas83"
                ame_reaction_1 = data_dir / "rct1.mas83"
                ame_reaction_2 = data_dir / "rct2.mas83"
            case 1993:
                ame_mass = data_dir / "mass_exp.mas93"
                ame_reaction_1 = data_dir / "rct1_exp.mas93"
                ame_reaction_2 = data_dir / "rct2_exp.mas93"
            case 1995:
                ame_mass = data_dir / "mass_exp.mas95"
                ame_reaction_1 = data_dir / "rct1_exp.mas95"
                ame_reaction_2 = data_dir / "rct2_exp.mas95"
            case 2003:
                ame_mass = data_dir / "mass.mas03"
                ame_reaction_1 = data_dir / "rct1.mas03"
                ame_reaction_2 = data_dir / "rct2.mas03"
            case 2012:
                ame_mass = data_dir / "mass.mas12"
                ame_reaction_1 = data_dir / "rct1.mas12"
                ame_reaction_2 = data_dir / "rct2.mas12"
            case 2016:
                ame_mass = data_dir / "mass16.txt"
                ame_reaction_1 = data_dir / "rct1-16.txt"
                ame_reaction_2 = data_dir / "rct2-16.txt"
            case 2020:
                ame_mass = data_dir / "mass.mas20"
                ame_reaction_1 = data_dir / "rct1.mas20"
                ame_reaction_2 = data_dir / "rct2.mas20"

        return ame_mass, ame_reaction_1, ame_reaction_2

    def _parse_nubase_data(self, year: int) -> pd.DataFrame:
        """Get the NUBASE for the given year as a pandas.DataFrame."""
        return NUBASEParser(self._get_nubase_datafile(year), year).read_file()

    def _parse_ame_data(self, year: int) -> pd.DataFrame:
        """Combine all the AME files from the given year into a pandas.DataFrame."""
        ame_mass, ame_reaction_1, ame_reaction_2 = self._get_ame_datafiles(year)

        ame_mass_df = AMEMassParser(ame_mass, year).read_file()

        # Merge all 3 of the AME files/data frames into one
        common_columns = ["A", "Z", "N", "TableYear", "Symbol"]
        temp_df = ame_mass_df.merge(AMEReactionParserOne(ame_reaction_1, year).read_file(), on=common_columns)
        return temp_df.merge(AMEReactionParserTwo(ame_reaction_2, year).read_file(), on=common_columns)

    def _combine_all_data(self) -> pd.DataFrame:
        """Combine all NUBASE and AME data into a single pandas DataFrame."""
        common_columns = ["A", "Z", "N", "TableYear", "Symbol"]
        df = pd.merge(self.ame, self.nubase, on=common_columns, how="outer")

        df["NUBASERelativeError"] = abs(df["NUBASEMassExcessError"] / df["NUBASEMassExcess"])
        df["AMERelativeError"] = abs(df["AMEMassExcessError"] / df["AMEMassExcess"])

        # 12C has a 0.0 +/ 0.0 mass excess by definition so calculating relative error -> NaN
        # Set the value to 0.0 as that's what it is
        df.loc[(df.Symbol == "C") & (df.A == 12), "NUBASERelativeError"] = 0.0
        df.loc[(df.Symbol == "C") & (df.A == 12), "AMERelativeError"] = 0.0

        # 198Au has a typo in it's decay mode in the 2012 table. It is recorded as '-'
        df.loc[(df.A == 198) & (df.Z == 79) & (df.TableYear == 2012), "Decay"] = "B-"

        return df

    def _do_indexing(self) -> None:
        """
        Set the index of the dataframe to the table year. This is done in place so nothing is returned.

        param: Nothing

        :return: Nothing
        """
        self.nubase.set_index("TableYear", inplace=True)
        self.ame.set_index("TableYear", inplace=True)
        self.full_data.set_index("TableYear", inplace=True)
